/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *   
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package xgi.ut.dsl.utl;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.beanutils.MethodUtils;
import org.mockito.cglib.proxy.Enhancer;

import xgi.ut.dsl.annot.ConversionMethod;
/**
 * This class provides a convenient workaround to avoid the fact that 
 * scala arrays are not automatically expanded to vararg when it comes
 * to java.lang.reflect.Method invocation
 */
public class ReflectionUtil {
	/** This field holds a cache of all the fields of all the previously inspected classes */
	private static final Map<Class<?>, Map<String, Field>> CACHE = new HashMap<Class<?>, Map<String,Field>>();
	
	/** 
	 * invokes the method identified by name on object target with the given parameters 
	 * and assigns it a mock result.
	 * @param target the object on which to invoke the method
	 * @param name the name of the method to invoke
	 * @param params the method parameters
	 * @param result the mock result to assign to the fn call
	 * @return the result of the fn call
	 * @throws NoSuchMethodException in case the method does not exists in the class
	 * @throws IllegalAccessException in case the method is not accessible
	 * @throws InvocationTargetException in case the method cannot be called
	 */
	public static final Object invoke(Object target, String name,
			List<?> params, Object result) throws NoSuchMethodException,
			IllegalAccessException, InvocationTargetException {
		Object invocation = null;
		Class<?> clazz = target.getClass();
		// was original class a stub ? 
		boolean wasOriginalClassAStub = Enhancer.isEnhanced(clazz);
		
		// retrieve the method
		// stub the params
		Class<?>[] paramTypes = new Class[params.size()];
		for (int i = 0; i < paramTypes.length; i++) {
			paramTypes[i] = params.get(i).getClass();
		}
		Method m = MethodUtils.getAccessibleMethod(clazz, name, paramTypes);
		
		if(m==null)throw new RuntimeException(name+Arrays.asList(paramTypes)+" not found in "+clazz);
		
		// Invoke the method
		
		try {
			invocation = m.invoke(target, params.toArray());
		} catch (Throwable e) {
			// swallow
		}
		// if it was a stub that returned null, then mock the result
		if (wasOriginalClassAStub) {
			Object fnRes = result;
			if (fnRes == null) {
				fnRes = mock(m.getReturnType());
			}
			try {
				// record the call in mockito
				when(m.invoke(target, params.toArray())).thenReturn(fnRes);
			} catch (Exception e) {
				e.printStackTrace();
			}
			invocation = fnRes;
		}
		return invocation;
	}
	
	/**
	 * This method provides a convenient way to invoke the conversion method of a conversion class
	 * @param text the text that needs to be converted (parsed)
	 * @param conversionClass the class that contains one conversion method (public static annotated with @ConversionMethod)
	 * @return the object that is represented by text
	 * 
	 * @throws IllegalArgumentException in case there is a mismatch btw the given arguments and that of the conversion method
	 */
	public static final Object invokeConversion(String text, Class<?> conversionClass) throws IllegalArgumentException {
		Method[] methods = conversionClass.getMethods();
		for(Method m : methods){
			ConversionMethod annotation = m.getAnnotation(ConversionMethod.class);
			if(annotation!=null){
				if(!Modifier.isStatic(m.getModifiers()))throw new RuntimeException("The conversion method must be static");
				if(!Modifier.isPublic(m.getModifiers()))throw new RuntimeException("The conversion method must be public");
				try {
					return m.invoke(null, text);
				} catch (IllegalAccessException e) {
					//swallow : it cannot happen
				} catch (InvocationTargetException e) {
					//swallow : it cannot happen
				}
			}
		}
		throw new RuntimeException("The parser class "+conversionClass.getCanonicalName()+" contains no public static method annotated with @ConversionMethod");
	}
	
	/**
	 * Retrieves all the public + private (declared + inherited) fields of a given class
	 * 
	 * Note: as a side effect, this method will populate the CACHE field of this class.
	 * 
	 * @param clazz the class to inspect
	 * @return a map with the key being the field name and the value the field itself.
	 */
	public static Map<String, Field> getAllFields(Class<?> clazz){
		Map<String, Field> results = CACHE.get(clazz);
		if(results!=null){
			return results;
		}
		if(clazz.equals(Object.class)){
			return new HashMap<String, Field>();
		}
		Class<?> parent = clazz.getSuperclass();
		results = new HashMap<String, Field>(getAllFields(parent));
		for(Field f : clazz.getDeclaredFields()){
			results.put(f.getName(), f);
		}
		CACHE.put(clazz, results);
		return results;
	}
}
